import functools
import json
import traceback
from concurrent.futures.thread import ThreadPoolExecutor
from threading import Lock
from typing import Any, Callable, Dict

from tornado.concurrent import run_on_executor
from tornado.ioloop import IOLoop
from tornado.web import HTTPError, RequestHandler

from bgmi.front.base import BaseHandler
from bgmi.lib.controllers import add, cal, cfg, delete, filter_, mark, search, status_, update
from bgmi.lib.download import download_prepare

ACTION_AUTH = "auth"
ACTION_STATUS = "status"


def auth_(token: str = "") -> Dict[str, str]:
    return {"status": "success" if token == cfg.http.admin_token else "error"}


API_MAP_POST: Dict[str, Callable] = {
    "add": add,
    "delete": delete,
    "search": search,
    "download": download_prepare,
    "auth": auth_,
    "mark": mark,
    "status": status_,
    "filter": filter_,
}

API_MAP_GET = {
    "cal": lambda: {"data": cal()},
    "config": lambda: {"data": json.loads(cfg.json())},
}

NO_AUTH_ACTION = ("cal", ACTION_AUTH)


def auth(f):  # type: ignore
    @functools.wraps(f)
    def wrapped(self: RequestHandler, *args: Any, **kwargs: Any) -> Any:
        if kwargs.get("action", None) in NO_AUTH_ACTION:
            return f(self, *args, **kwargs)

        token = self.request.headers.get("bgmi-token")
        if token == cfg.http.admin_token:
            return f(self, *args, **kwargs)
        else:
            # HTTPError will be except in `BaseHandler.write_error`
            raise HTTPError(401)

    return wrapped


class AdminApiHandler(BaseHandler):
    @auth
    def get(self, action: str) -> None:
        try:
            result = API_MAP_GET[action]()
        except KeyError:
            raise HTTPError(404)
        except Exception:
            traceback.print_exc()
            raise HTTPError(400)
        self.finish(self.jsonify(**result))

    @auth
    def post(self, action: str) -> None:
        data = self.get_json()

        try:
            result = API_MAP_POST[action](**data)
            if result["status"] == "error":
                raise HTTPError(400)
        except HTTPError:
            raise HTTPError(400)
        except KeyError:
            raise HTTPError(404)
        except Exception:
            traceback.print_exc()
            raise HTTPError(500)

        resp = self.jsonify(**result)
        self.finish(resp)


class UpdateHandler(BaseHandler):
    executor = ThreadPoolExecutor(2)  # pylint: disable=consider-using-with
    lock = Lock()

    @auth
    async def post(self) -> None:
        if not self.lock.locked():
            with self.lock:
                # 这样将在下一轮事件循环执行self.sleep
                IOLoop.instance().add_callback(self.hard_task)
                await self.finish(self.jsonify(message="start updating"))
        else:
            await self.finish(self.jsonify(message="previous updating has not finish"))

    @run_on_executor
    def hard_task(self) -> None:
        print("start work")
        data = self.get_json()

        name = data.get("name", "")
        download = data.get("download", [])

        if not download:
            download = None
        update(name, download)
        self.lock.release()



import os
from pathlib import Path
from shutil import copy
from typing import List, Type

from bgmi import __version__
from bgmi.config import BGMI_PATH, IS_WINDOWS, cfg
from bgmi.lib import models
from bgmi.lib.models import NeoDB
from bgmi.utils import print_error, print_info, print_success, print_warning


def install_crontab() -> None:
    print_info("Installing crontab job")
    if IS_WINDOWS:
        copy(os.path.join(os.path.dirname(__file__), "others/cron.vbs"), BGMI_PATH)
        os.system(
            'schtasks /Create /SC HOURLY /TN "bgmi updater" /TR "{}"  /IT /F'.format(
                os.path.join(BGMI_PATH, "cron.vbs")
            )
        )
    else:
        path = os.path.join(os.path.dirname(__file__), "others/crontab.sh")
        os.system(f"bash '{path}'")


def create_dir() -> None:
    path_to_create: List[Path] = [
        BGMI_PATH,
        cfg.save_path,
        cfg.tmp_path,
        cfg.script_path,
        cfg.tools_path,
        cfg.front_static_path,
    ]

    if not os.environ.get("HOME", os.environ.get("USERPROFILE", "")):
        print_warning("$HOME not set, use '/tmp/'")

    # bgmi home dir
    try:
        for path in path_to_create:
            if not os.path.exists(path):
                os.makedirs(path, exist_ok=True)
                print_success(f"{path} created successfully")
        OLD = os.path.join(BGMI_PATH, "old")
        # create OLD if not exist oninstall
        if not os.path.exists(OLD):
            with open(OLD, "w", encoding="utf8") as f:
                f.write(__version__)
    except OSError as e:
        print_error(f"Error: {str(e)}")


def init_db() -> None:
    tables: List[Type[NeoDB]] = [
        models.Scripts,
        models.Bangumi,
        models.Followed,
        models.Subtitle,
        models.Filter,
        models.Download,
    ]

    for t in tables:
        t.create_table()



import datetime
import itertools
import os
import sys
from operator import itemgetter
from typing import List, Mapping, Optional, Tuple

import click
import pydantic
import tomlkit
import wcwidth
from loguru import logger
from tornado import template

from bgmi import __version__
from bgmi.config import BGMI_PATH, CONFIG_FILE_PATH, Config, cfg, print_config, write_default_config
from bgmi.lib import controllers as ctl
from bgmi.lib.constants import BANGUMI_UPDATE_TIME, SPACIAL_APPEND_CHARS, SPACIAL_REMOVE_CHARS, SUPPORT_WEBSITE
from bgmi.lib.download import download_prepare
from bgmi.lib.fetch import website
from bgmi.lib.models import STATUS_DELETED, STATUS_FOLLOWED, STATUS_UPDATED, Bangumi, Filter, Followed, Subtitle
from bgmi.lib.update import update_database
from bgmi.script import ScriptRunner
from bgmi.setup import create_dir, init_db, install_crontab
from bgmi.utils import (
    COLOR_END,
    GREEN,
    RED,
    YELLOW,
    check_update,
    get_terminal_col,
    get_web_admin,
    print_error,
    print_info,
    print_success,
    print_version,
    print_warning,
)

__all__ = ["main_for_test", "main", "print_success"]


def main() -> None:
    logger.remove()
    logger.add(
        sys.stderr, format="<blue>{time:YYYY-MM-DD HH:mm:ss}</blue> {level:7} | <level>{message}</level>", level="INFO"
    )
    logger.add(cfg.log_path.parent.joinpath("{time:YYYY-MM-DD}.log"), format="{time} {level} {message}", level="INFO")

    cli.main(prog_name="bgmi")


def main_for_test(args: Optional[List[str]] = None) -> None:
    cli.main(args=args, prog_name="bgmi", standalone_mode=False)


@click.group(name="bgmi")
@click.version_option(__version__, package_name="bgmi", prog_name="bgmi", message=print_version())
@click.pass_context
def cli(ctx: click.Context) -> None:
    if ctx.command not in ["install", "upgrade", "complete"]:
        check_update()


@cli.command(help="Install BGmi and frontend")
def install() -> None:
    need_to_init = False
    if not os.path.exists(BGMI_PATH):
        need_to_init = True
        print_warning(f"BGMI_PATH {BGMI_PATH} does not exist, installing")

    create_dir()
    init_db()
    if need_to_init:
        install_crontab()

    write_default_config()
    update_database()
    get_web_admin(method="install")


@cli.command(help="upgrade from previous version")
def upgrade() -> None:
    create_dir()
    update_database()
    check_update()


@cli.command(
    help="Select date source bangumi_moe or mikan_project",
)
@click.argument("bangumi_source", required=True, type=click.Choice([x["id"] for x in SUPPORT_WEBSITE]))
def source(bangumi_source: str) -> None:
    result = ctl.source(data_source=bangumi_source)
    globals()["print_{}".format(result["status"])](result["message"])


@cli.group()
def config() -> None:
    ...


@config.command("print")
def config_print() -> None:
    print(print_config())


@config.command("set")
@click.argument("keys", nargs=-1)
@click.option("--value", required=True)
def config_set(keys: List[str], value: str) -> None:
    doc = tomlkit.loads(CONFIG_FILE_PATH.read_text(encoding="utf-8"))

    if keys[0] == "source":
        print_error("you can't change source with this command, use `bgmi source ...`", stop=True)

    res = doc
    for key in keys[:-1]:
        res = res.setdefault(key, {})
        if not isinstance(res, Mapping):
            print_error(f"value of key '{key}' is not object, can't update its attribute")

    res[keys[-1]] = value

    try:
        Config.parse_obj(doc)
    except pydantic.ValidationError as e:
        print(e)
        print("config is not valid after change, won't write to config file")
        return

    CONFIG_FILE_PATH.write_text(tomlkit.dumps(doc))


@config.command("get")
@click.argument("keys", nargs=-1)
def config_get(keys: List[str]) -> None:
    doc = tomlkit.loads(CONFIG_FILE_PATH.read_text(encoding="utf-8"))
    res = doc

    for key in keys:
        res = doc.get(key, {})

    print("config", ".".join(keys), res)


@cli.command(help="Search torrents from data source by keyword")
@click.argument("keyword")
@click.option("--count", type=int, help="The max page count of search result.")
@click.option("--regex-filter", "regex", help="Regular expression filter of title.")
@click.option("--download", is_flag=True, show_default=True, default=False, type=bool, help="Download search result.")
@click.option("--dupe", is_flag=True, show_default=True, default=False, type=bool, help="Show duplicated episode")
@click.option("--min-episode", "min_episode", type=int, help="Minimum episode filter of title.")
@click.option("--max-episode", "max_episode", type=int, help="Maximum episode filter of title.")
@click.option(
    "--tag", is_flag=True, show_default=True, default=False, help="Use tag to search (if data source supported)."
)
@click.option("--subtitle", help="Subtitle group filter of title (Need --tag enabled)")
def search(
    keyword: str,
    count: int,
    regex: str,
    download: bool,
    dupe: bool,
    min_episode: int,
    max_episode: int,
    tag: bool,
    subtitle: str,
) -> None:
    result = ctl.search(
        keyword=keyword,
        count=count,
        regex=regex,
        dupe=dupe,
        min_episode=min_episode,
        max_episode=max_episode,
        tag=tag,
        subtitle=subtitle,
    )
    if result["status"] != "success":
        globals()["print_{}".format(result["status"])](result["message"])
    data = result["data"]
    for i in data:
        print(i.title)
    if download:
        download_prepare(data)


@cli.command("mark")
@click.argument("name", required=True)
@click.argument("episode", type=int, required=True)
def mark(name: str, episode: int) -> None:
    result = ctl.mark(name=name, episode=episode)
    globals()["print_{}".format(result["status"])](result["message"])


@cli.command()
@click.argument("names", nargs=-1)
@click.option(
    "--episode",
    type=int,
    help="add bangumi and mark it as specified episode",
)
def add(names: List[str], episode: Optional[int]) -> None:
    """
    subscribe bangumi

    names: list of bangumi names to subscribe
    """
    for name in names:
        result = ctl.add(name=name, episode=episode)
        globals()["print_{}".format(result["status"])](result["message"])


@cli.command()
@click.argument("name", nargs=-1)
@click.option(
    "--clear-all",
    "clear",
    is_flag=True,
    default=False,
    help="Clear all the subscriptions, name will be ignored If you provide this flag",
)
@click.option("--yes", is_flag=True, default=False, help="No confirmation")
def delete(name: List[str], clear: bool, yes: bool) -> None:
    """
    name: list of bangumi names to unsubscribe
    """
    if clear:
        ctl.delete("", clear_all=clear, batch=yes)
    else:
        for bangumi_name in name:
            result = ctl.delete(name=bangumi_name)
            globals()["print_{}".format(result["status"])](result["message"])


@cli.command("list", help="list subscribed bangumi")
def list_command() -> None:
    result = ctl.list_()
    print(result["message"])


@cli.command("filter", help="set bangumi episode filters")
@click.argument("name", required=True)
@click.option("--subtitle", help='Subtitle group name, split by ",".')
@click.option(
    "--include",
    help='Filter by keywords which in the title, split by ",".',
)
@click.option("--exclude", help='Filter by keywords which not int the title, split by ",".')
@click.option("--regex", help="Filter by regular expression")
def filter_cmd(
    name: str,
    subtitle: Optional[str],
    regex: Optional[str],
    include: Optional[str],
    exclude: Optional[str],
) -> None:
    """
    name: bangumi name to update filter
    """
    result = ctl.filter_(
        name=name,
        subtitle=subtitle,
        include=include,
        exclude=exclude,
        regex=regex,
    )
    if "data" not in result:
        globals()["print_{}".format(result["status"])](result["message"])
    else:
        print_info("Usable subtitle group: {}".format(", ".join(result["data"]["subtitle_group"])))
        followed_filter_obj = Filter.get(bangumi_name=result["data"]["name"])
        print_filter(followed_filter_obj)


def print_filter(followed_filter_obj: Filter) -> None:
    print(
        "Followed subtitle group: {}".format(
            ", ".join(x["name"] for x in Subtitle.get_subtitle_by_id(followed_filter_obj.subtitle.split(", ")))
            if followed_filter_obj.subtitle
            else "None"
        )
    )
    print(f"Include keywords: {followed_filter_obj.include}")
    print(f"Exclude keywords: {followed_filter_obj.exclude}")
    print(f"Regular expression: {followed_filter_obj.regex}")


@cli.command("cal")
@click.option(
    "-f",
    "--force-update",
    "force_update",
    is_flag=True,
    show_default=True,
    default=False,
    type=bool,
    help="get latest bangumi calendar",
)
@click.option(
    "--today",
    "today",
    is_flag=True,
    show_default=True,
    default=False,
    type=bool,
    help="show bangumi calendar for today.",
)
@click.option(
    "--download-cover",
    "download_cover",
    is_flag=True,
    show_default=True,
    default=False,
    type=bool,
    help="download the cover to local",
)
def calendar(force_update: bool, today: bool, download_cover: bool) -> None:
    runner = ScriptRunner()
    cover: Optional[List[str]] = None

    if download_cover:
        cover = runner.get_download_cover()

    weekly_list = ctl.cal(force_update=force_update, cover=cover)

    def shift(seq: Tuple[str, ...], n: int) -> Tuple[str, ...]:
        n %= len(seq)
        return seq[n:] + seq[:n]

    order_without_unknown = BANGUMI_UPDATE_TIME[:-1]
    if today:
        weekday_order = (order_without_unknown[datetime.datetime.today().weekday()],)  # type: Tuple[str, ...]
    else:
        weekday_order = shift(order_without_unknown, datetime.datetime.today().weekday())

    col = max(wcwidth.wcswidth(bangumi["name"]) for value in weekly_list.values() for bangumi in value)
    env_columns = col if os.environ.get("TRAVIS_CI", False) else get_terminal_col()

    if env_columns < col:
        print_warning("terminal window is too small.")
        env_columns = col

    row = int(env_columns / col if env_columns / col <= 3 else 3)

    def print_line() -> None:
        num = col - 3
        split = "-" * num + "   "
        print(split * row)

    for weekday in weekday_order + ("Unknown",):
        if weekly_list[weekday.lower()]:
            print(
                "{}{}. {}".format(
                    GREEN,
                    weekday if not today else f"Bangumi Schedule for Today ({weekday})",
                    COLOR_END,
                ),
                end="",
            )
            print()
            print_line()
            for i, bangumi in enumerate(weekly_list[weekday.lower()]):
                if bangumi["status"] in (STATUS_UPDATED, STATUS_FOLLOWED) and "episode" in bangumi:
                    bangumi["name"] = "{}({:d})".format(bangumi["name"], bangumi["episode"])

                width = wcwidth.wcswidth(bangumi["name"])
                space_count = col - 2 - width

                for s in SPACIAL_APPEND_CHARS:
                    if s in bangumi["name"]:
                        space_count += bangumi["name"].count(s)

                for s in SPACIAL_REMOVE_CHARS:
                    if s in bangumi["name"]:
                        space_count -= bangumi["name"].count(s)

                if bangumi["status"] == STATUS_FOLLOWED:
                    bangumi["name"] = "{}{}{}".format(YELLOW, bangumi["name"], COLOR_END)

                if bangumi["status"] == STATUS_UPDATED:
                    bangumi["name"] = "{}{}{}".format(GREEN, bangumi["name"], COLOR_END)
                try:
                    print(" " + bangumi["name"], " " * space_count, end="")
                except UnicodeEncodeError:
                    continue

                if (i + 1) % row == 0 or i + 1 == len(weekly_list[weekday.lower()]):
                    print()
            print()


@cli.command("fetch")
@click.argument("name")
@click.option(
    "--not-ignore", "not_ignore", is_flag=True, help="Do not ignore the old bangumi detail rows (3 month ago)"
)
def fetch(name: str, not_ignore: bool) -> None:
    """
    name: bangumi name to fetch
    """

    try:
        bangumi_obj = Bangumi.get(name=name)
    except Bangumi.DoesNotExist:
        print_error(f"Bangumi {name} not exist", stop=True)
        return

    try:
        Followed.get(bangumi_name=bangumi_obj.name)
    except Followed.DoesNotExist:
        print_error(f"Bangumi {name} is not followed")
        return

    followed_filter_obj = Filter.get(bangumi_name=name)
    print_filter(followed_filter_obj)

    print_info(f"Fetch bangumi {bangumi_obj.name} ...")
    _, data = website.get_maximum_episode(bangumi_obj, ignore_old_row=not bool(not_ignore))

    if not data:
        print_warning("Nothing.")

    max_episode = max(i.episode for i in data)
    digest = len(str(max_episode))

    for i in data:
        episode = str(i.episode).rjust(digest)
        print(f"{episode} | {i.title}")


@cli.command("update", help="Update bangumi calendar and subscribed bangumi episode.")
@click.argument(
    "names",
    nargs=-1,
)
@click.option(
    "-d", "--download", is_flag=True, default=False, help="Download specified episode of the bangumi when updated"
)
@click.option(
    "--not-ignore", "not_ignore", is_flag=True, help="Do not ignore the old bangumi detail rows (3 month ago)"
)
def update(names: List[str], download: bool, not_ignore: bool) -> None:
    """
    name: optional bangumi name list you want to update
    """
    ctl.update(names, download=download, not_ignore=not_ignore)


@cli.command("gen")
@click.argument("tpl", type=click.Choice(["nginx.conf"]))
@click.option("--server-name", "server_name")
def generate_config(tpl: str, server_name: str) -> None:
    template_file_path = os.path.join(os.path.dirname(__file__), "others", "nginx.conf")

    with open(template_file_path, encoding="utf8") as template_file:
        shell_template = template.Template(template_file.read(), autoescape="")

    template_with_content = shell_template.generate(
        server_name=server_name,
        os_sep=os.sep,
        front_static_path=str(cfg.front_static_path.as_posix()),
        save_path=str(cfg.save_path.as_posix()),
    )

    print(template_with_content.decode("utf-8"))


@cli.command("history", help="list your history of following bangumi")
def history() -> None:
    m = (
        "January",
        "February",
        "March",
        "April",
        "May",
        "June",
        "July",
        "August",
        "September",
        "October",
        "November",
        "December",
    )
    data = Followed.select(Followed).order_by(Followed.updated_time.asc())
    bangumi_data = Bangumi.get_updating_bangumi()
    year = None
    month = None

    updating_bangumi = list(map(itemgetter("name"), itertools.chain(*bangumi_data.values())))

    print("Bangumi Timeline")
    for i in data:
        if i.status == STATUS_DELETED:
            slogan = "ABANDON"
            color = RED
        else:
            if i.bangumi_name in updating_bangumi:
                slogan = "FOLLOWING"
                color = YELLOW
            else:
                slogan = "FINISHED"
                color = GREEN

        if not i.updated_time:
            date = datetime.datetime.fromtimestamp(0)
        else:
            date = datetime.datetime.fromtimestamp(int(i.updated_time))

        if date.year != 1970:
            if date.year != year:
                print(f"{GREEN}{str(date.year)}{COLOR_END}")
                year = date.year

            if date.year == year and date.month != month:
                print(f"  |\n  |--- {YELLOW}{m[date.month - 1]}{COLOR_END}\n  |      |")
                month = date.month

            print(f"  |      |--- [{color}{slogan:<9}{COLOR_END}] ({i.episode:<2}) {i.bangumi_name}")



from operator import attrgetter
from typing import List, Optional

from pydantic import BaseModel, validator

from bgmi.lib.constants import BANGUMI_UPDATE_TIME


class Episode(BaseModel):
    title: str
    download: str
    episode: int = 0
    time: int = 0
    subtitle_group: Optional[str]
    name: str = ""

    @staticmethod
    def remove_duplicated_bangumi(result: List["Episode"]) -> List["Episode"]:
        ret = []
        episodes = list({i.episode for i in result})
        for i in result:
            if i.episode in episodes:
                ret.append(i)
                del episodes[episodes.index(i.episode)]

        return ret

    def contains_any_words(self, keywords: List[str]) -> bool:
        """Keywords should be converted to low case after passed to this function."""
        title = self.title.lower()
        return any(t in title for t in keywords)


class SubtitleGroup(BaseModel):
    id: str
    name: str


class WebsiteBangumi(BaseModel):
    keyword: str
    update_time: str = "Unknown"
    name: str = ""
    status: int = 0
    subtitle_group: List[SubtitleGroup] = []
    cover: str = ""
    episodes: List[Episode] = []

    @property
    def max_episode(self) -> int:
        return max(self.episodes, key=attrgetter("episode")).episode

    @validator("update_time")
    def validate_update_time(cls, v: str) -> str:
        # pylint: disable=no-self-argument
        assert v in BANGUMI_UPDATE_TIME, f"update time can be only one of {BANGUMI_UPDATE_TIME}"
        return v



import os
import re
import time
import urllib.parse
from typing import List, Optional

import requests
from bs4 import BeautifulSoup
from loguru import logger

from bgmi.config import cfg
from bgmi.session import session
from bgmi.utils import print_error
from bgmi.website.base import BaseWebsite
from bgmi.website.model import Episode, SubtitleGroup, WebsiteBangumi

base_url = cfg.share_dmhy_url


def fetch_url(url, **kwargs):
    ret = None
    try:
        ret = session.get(url, timeout=60, **kwargs).text
    except requests.ConnectionError:
        logger.error("Create connection to {}... failed", base_url)
        print_error("Check internet connection or try to set a DMHY mirror site with share_dmhy_url in config")

    return ret


def parse_bangumi_with_week_days(content, update_time, array_name) -> List[WebsiteBangumi]:
    content = re.sub("'\\+encodeURIComponent\\('(.*?)'\\)\\+'", "\\1", content)
    content = re.sub("encodeURIComponent\\('(.*?)'\\)", "'URLE#\\1'", content)
    r = re.compile(array_name + "\\.push\\(\\['(.*?)','(.*?)','(.*?)','(.*?)','(.*?)'\\]\\)")
    ret = r.findall(content)

    bangumi_list = []

    for bangumi_row in ret:
        (cover_url, name, keyword, subtitle_raw, _) = bangumi_row
        if keyword.startswith("URLE#"):
            keyword = urllib.parse.quote(keyword[5:])
        bangumi = WebsiteBangumi(keyword=keyword)
        cover = re.findall("(/images/.*)$", cover_url)[0]

        bs = BeautifulSoup(subtitle_raw, "html.parser")
        a_list = bs.find_all("a")

        for a in a_list:
            subtitle_group_name = a.get_text(strip=True)
            subtitle_group_id_raw = re.findall("team_id%3A(.+)$", a["href"])

            if (len(subtitle_group_id_raw) == 0) or subtitle_group_name == "":
                continue

            subtitle_group_id = subtitle_group_id_raw[0]

            bangumi.subtitle_group.append(SubtitleGroup(id=subtitle_group_id, name=subtitle_group_name))

        bangumi.name = name
        bangumi.update_time = update_time
        bangumi.keyword = keyword
        bangumi.cover = base_url + cover

        # append to bangumi_list
        bangumi_list.append(bangumi)

    return bangumi_list


def parse_subtitle_list(content):
    subtitle_list = []

    bs = BeautifulSoup(content, "html.parser")
    li_list = bs.find_all("li", {"class": "team-item"})

    for li in li_list:
        subtitle_group_name = li.span.a.get("title")
        subtitle_group_id_raw = re.findall(r"team_id\/(.+)$", li.span.a.get("href"))

        if (len(subtitle_group_id_raw) == 0) or subtitle_group_name == "":
            continue

        subtitle_group_id = subtitle_group_id_raw[0]
        # append to subtitle_list
        subtitle_list.append({"id": subtitle_group_id, "name": subtitle_group_name})

    return subtitle_list


def unique_subtitle_list(raw_list):
    ret = []
    id_list = list({i["id"] for i in raw_list})
    for row in raw_list:
        if row["id"] in id_list:
            ret.append(row)
            del id_list[id_list.index(row["id"])]

    return ret


class DmhySource(BaseWebsite):
    def search_by_keyword(self, keyword, count=None):
        """
        return a list of dict with at least 4 key: download, name, title, episode
        example:
        ```
            [
                {
                    'name':"路人女主的养成方法",
                    'download': 'magnet:?xt=urn:btih:what ever',
                    'title': "[澄空学园] 路人女主的养成方法 第12话 MP4 720p  完",
                    'episode': 12
                },
            ]
        ```
        :param keyword: search key word
        :type keyword: str
        :param count: how many page to fetch from website
        :type count: int

        :return: list of episode search result
        :rtype: list[dict]
        """
        if count is None:
            count = 3

        result = []
        search_url = base_url + "/topics/list/"
        for i in range(count):
            params = {"keyword": keyword, "page": i + 1}

            if os.environ.get("DEBUG", False):  # pragma: no cover
                print(search_url, params)

            r = fetch_url(search_url, params=params)
            bs = BeautifulSoup(r, "html.parser")

            table = bs.find("table", {"id": "topic_list"})
            if table is None:
                break
            tr_list = table.tbody.find_all("tr", {"class": ""})
            for tr in tr_list:
                td_list = tr.find_all("td")

                if td_list[1].a["class"][0] != "sort-2":
                    continue

                time_string = td_list[0].span.string
                name = keyword
                title = td_list[2].find("a", {"target": "_blank"}).get_text(strip=True)
                download = td_list[3].a["href"]
                episode = self.parse_episode(title)
                t = int(time.mktime(time.strptime(time_string, "%Y/%m/%d %H:%M")))

                result.append(
                    Episode(
                        name=name,
                        title=title,
                        download=download,
                        episode=episode,
                        time=t,
                    )
                )

        return result

    def fetch_bangumi_calendar(self):
        week_days_mapping = [
            ("Sun", "sunarray"),
            ("Mon", "monarray"),
            ("Tue", "tuearray"),
            ("Wed", "wedarray"),
            ("Thu", "thuarray"),
            ("Fri", "friarray"),
            ("Sat", "satarray"),
        ]

        bangumi_list = []

        url = base_url + "/cms/page/name/programme.html"

        r = fetch_url(url)

        for update_time, array_name in week_days_mapping:
            b_list = parse_bangumi_with_week_days(r, update_time, array_name)
            bangumi_list.extend(b_list)

        return bangumi_list

    def search_by_tag(self, tag: str, subtitle: Optional[str] = None, count: Optional[int] = None) -> List[Episode]:
        print_error("dmhy not support search by tag")
        return []

    def fetch_episode_of_bangumi(self, bangumi_id, max_page=cfg.max_path, subtitle_list=None):
        """
        get all episode by bangumi id
        example
        ```
            [
                {
                    "download": "magnet:?xt=urn:btih:e43b3b6b53dd9fd6af1199e112d3c7ff15",
                    "subtitle_group": "58a9c1c9f5dc363606ab42ec",
                    "title": "【喵萌奶茶屋】★七月新番★[来自深渊/Made in Abyss][07][GB][720P]",
                    "episode": 0,
                    "time": 1503301292
                },
            ]
        ```

        :param bangumi_id: bangumi_id
        :param subtitle_list: list of subtitle group
        :type subtitle_list: list
        :param max_page: how many page you want to crawl if there is no subtitle list
        :type max_page: int
        :return: list of bangumi
        :rtype: list[dict]
        """
        result = []
        keyword = bangumi_id
        search_url = base_url + "/topics/list/"
        for i in range(max_page):
            url = search_url + "?keyword=" + keyword + "&page=" + str(i + 1)

            if os.environ.get("DEBUG", False):  # pragma: no cover
                print(url)

            r = fetch_url(url)
            bs = BeautifulSoup(r, "html.parser")

            table = bs.find("table", {"id": "topic_list"})
            if table is None:
                break
            tr_list = table.tbody.find_all("tr", {"class": ""})
            for tr in tr_list:
                td_list = tr.find_all("td")

                if td_list[1].a["class"][0] != "sort-2":
                    continue

                time_string = td_list[0].span.string
                name = keyword
                title = td_list[2].find("a", {"target": "_blank"}).get_text(strip=True)
                download = td_list[3].a["href"]
                episode = self.parse_episode(title)
                t = int(time.mktime(time.strptime(time_string, "%Y/%m/%d %H:%M")))
                subtitle_group = ""

                tag_list = td_list[2].find_all("span", {"class": "tag"})

                for tag in tag_list:
                    href = tag.a.get("href")
                    if href is None:
                        continue

                    team_id_raw = re.findall(r"team_id\/(.*)$", href)
                    if len(team_id_raw) == 0:
                        continue
                    subtitle_group = team_id_raw[0]

                if subtitle_list:
                    if subtitle_group not in subtitle_list:
                        continue

                if os.environ.get("DEBUG", False):  # pragma: no cover
                    print(name, title, subtitle_group, download, episode, t)

                result.append(
                    Episode(
                        title=title,
                        subtitle_group=subtitle_group,
                        download=download,
                        episode=episode,
                        time=t,
                    )
                )

        return result



import os
import time
from collections import defaultdict
from typing import Any, Dict, List, Optional, Tuple, Type, TypeVar

import peewee
from peewee import FixedCharField, IntegerField, TextField
from playhouse.shortcuts import model_to_dict

from bgmi.config import cfg
from bgmi.lib.constants import BANGUMI_UPDATE_TIME
from bgmi.utils import episode_filter_regex
from bgmi.website.model import Episode

# bangumi status
STATUS_UPDATING = 0
STATUS_END = 1
BANGUMI_STATUS = (STATUS_UPDATING, STATUS_END)

# subscription status
STATUS_DELETED = 0
STATUS_FOLLOWED = 1
STATUS_UPDATED = 2
FOLLOWED_STATUS = (STATUS_DELETED, STATUS_FOLLOWED, STATUS_UPDATED)

# download status
STATUS_NOT_DOWNLOAD = 0
STATUS_DOWNLOADING = 1
STATUS_DOWNLOADED = 2
DOWNLOAD_STATUS = (STATUS_NOT_DOWNLOAD, STATUS_DOWNLOADING, STATUS_DOWNLOADED)

DoesNotExist = peewee.DoesNotExist

db = peewee.SqliteDatabase(cfg.db_path)

if os.environ.get("DEV"):
    print(f"using database {cfg.db_path}")

_Cls = TypeVar("_Cls")


class NeoDB(peewee.Model):
    DoesNotExist: Type[peewee.DoesNotExist]

    class Meta:
        database = db

    @classmethod
    def get(cls: Type[_Cls], *query: Any, **filters: Any) -> _Cls:
        return super().get(*query, **filters)  # type: ignore

    @classmethod
    def get_or_create(cls: Type[_Cls], **kwargs: Any) -> Tuple[_Cls, bool]:
        return super().get_or_create(**kwargs)  # type: ignore


class Bangumi(NeoDB):
    id = IntegerField(primary_key=True)
    name = TextField(unique=True, null=False)
    subtitle_group = TextField(null=False)
    keyword = TextField()
    update_time = FixedCharField(5, null=False)
    cover = TextField()
    status = IntegerField(default=0)

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)

        update_time = kwargs.get("update_time", "").title()
        if update_time and update_time not in BANGUMI_UPDATE_TIME:
            raise ValueError(f"unexpected update time {update_time}")
        self.update_time = update_time
        if isinstance(kwargs.get("subtitle_group"), list):
            s = []
            for sub in kwargs["subtitle_group"]:
                if isinstance(sub, str):
                    s.append(sub)
                elif isinstance(sub, dict):
                    s.append(sub["id"])
                else:
                    s.append(sub.id)
            self.subtitle_group = ", ".join(sorted(s))

    @classmethod
    def delete_all(cls) -> None:
        un_updated_bangumi: List[Followed] = Followed.select().where(
            Followed.updated_time > (int(time.time()) - 2 * 7 * 24 * 3600)
        )
        if os.getenv("DEBUG"):  # pragma: no cover
            print("ignore updating bangumi", [x.bangumi_name for x in un_updated_bangumi])

        cls.update(status=STATUS_END).where(
            cls.name.not_in([x.bangumi_name for x in un_updated_bangumi])
        ).execute()  # do not mark updating bangumi as STATUS_END

    @classmethod
    def get_updating_bangumi(cls, status: Optional[int] = None, order: bool = True) -> Any:
        if status is None:
            data = (
                cls.select(Followed.status, Followed.episode, cls)
                .join(
                    Followed,
                    peewee.JOIN["LEFT_OUTER"],
                    on=(cls.name == Followed.bangumi_name),
                )
                .where(cls.status == STATUS_UPDATING)
                .dicts()
            )
        else:
            data = (
                cls.select(Followed.status, Followed.episode, cls)
                .join(
                    Followed,
                    peewee.JOIN["LEFT_OUTER"],
                    on=(cls.name == Followed.bangumi_name),
                )
                .where((cls.status == STATUS_UPDATING) & (Followed.status == status))
                .dicts()
            )

        if order:
            weekly_list = defaultdict(list)
            for bangumi_item in data:
                weekly_list[bangumi_item["update_time"].lower()].append(dict(bangumi_item))
        else:
            weekly_list = list(data)  # type: ignore

        return weekly_list

    @classmethod
    def fuzzy_get(cls, **filters: Any) -> "Bangumi":
        fuzzy_q = []
        raw_q = []
        for key, value in filters.items():
            raw_q.append(getattr(cls, key) == value)
            fuzzy_q.append(getattr(cls, key).contains(value))

        raw = list(cls.select().where(*raw_q))  # type: List[Bangumi]
        if raw:
            return raw[0]

        fuzzy = list(cls.select().where(*fuzzy_q))  # type: List[Bangumi]
        if fuzzy:
            return fuzzy[0]

        raise cls.DoesNotExist


class Followed(NeoDB):
    bangumi_name = TextField(unique=True)
    episode = IntegerField(null=True)
    status = IntegerField(null=True)
    updated_time = IntegerField(null=True)

    class Meta:
        database = db
        table_name = "followed"

    @classmethod
    def delete_followed(cls, batch: bool = True) -> bool:
        q = cls.delete()
        if not batch:
            if not input("[+] are you sure want to CLEAR ALL THE BANGUMI? (y/N): ") == "y":
                return False

        q.execute(None)
        return True

    @classmethod
    def get_all_followed(cls, status: int = STATUS_DELETED, bangumi_status: int = STATUS_UPDATING) -> List[dict]:
        join_cond = Bangumi.name == cls.bangumi_name
        d = (
            cls.select(Bangumi.name, Bangumi.update_time, Bangumi.cover, cls)
            .join(Bangumi, peewee.JOIN["LEFT_OUTER"], on=join_cond)
            .where((cls.status != status) & (Bangumi.status == bangumi_status))
            .order_by(cls.updated_time.desc())
            .dicts()
        )

        return list(d)


class Download(NeoDB):
    name = TextField(null=False)
    title = TextField(null=False)
    episode = IntegerField(default=0)
    download = TextField()
    status = IntegerField(default=0)

    @classmethod
    def get_all_downloads(cls, status: Optional[int] = None) -> List[dict]:
        if status is None:
            data = list(cls.select().order_by(cls.status))
        else:
            data = list(cls.select().where(cls.status == status).order_by(cls.status))

        for index, x in enumerate(data):
            data[index] = model_to_dict(x)
        return data

    def downloaded(self) -> None:
        self.status = STATUS_DOWNLOADED
        self.save()


class Filter(NeoDB):
    bangumi_name = TextField(unique=True)  # type: Optional[str]
    subtitle = TextField(null=True)  # type: Optional[str]
    include = TextField(null=True)  # type: Optional[str]
    exclude = TextField(null=True)  # type: Optional[str]
    regex = TextField(null=True)  # type: Optional[str]

    @property
    def subtitle_group_split(self) -> List[str]:
        if self.subtitle:
            # pylint:disable=no-member
            return [x.strip() for x in self.subtitle.split(",")]
        else:
            return []

    def apply_on_episodes(self, result: List[Episode]) -> List[Episode]:
        if self.include:
            # pylint:disable=no-member
            include_list = [s.strip().lower() for s in self.include.split(",")]
            result = [e for e in result if e.contains_any_words(include_list)]

        if cfg.enable_global_include_keywords:
            include_list = [s.strip().lower() for s in cfg.global_include_keywords]
            result = [e for e in result if e.contains_any_words(include_list)]

        if self.exclude:
            # pylint:disable=no-member
            exclude_list = [s.strip().lower() for s in self.exclude.split(",")]
            result = [e for e in result if not e.contains_any_words(exclude_list)]

        return episode_filter_regex(data=result, regex=self.regex)


class Subtitle(NeoDB):
    id = TextField(primary_key=True, unique=True)
    name = TextField()

    @classmethod
    def get_subtitle_by_id(cls, id_list: List[str]) -> List[Dict[str, str]]:
        data = list(cls.select().where(cls.id.in_(id_list)))
        for index, subtitle in enumerate(data):
            data[index] = model_to_dict(subtitle)
        return data

    @classmethod
    def get_subtitle_by_name(cls, name_list: List[str]) -> List[Dict[str, str]]:
        data = list(cls.select().where(cls.name.in_(name_list)))
        for index, subtitle in enumerate(data):
            data[index] = model_to_dict(subtitle)
        return data


class Scripts(NeoDB):
    bangumi_name = TextField(null=False, unique=True)
    episode = IntegerField(default=0)
    status = IntegerField(default=0)
    updated_time = IntegerField(default=0)


def recreate_source_relatively_table() -> None:
    table_to_drop = [
        Bangumi,
        Followed,
        Subtitle,
        Filter,
        Download,
    ]  # type: List[Type[NeoDB]]
    for table in table_to_drop:
        table.delete().execute()  # pylint: disable=no-value-for-parameter


def recreate_scripts_table() -> None:
    table_to_drop = [
        Scripts,
    ]  # type: List[Type[NeoDB]]
    for table in table_to_drop:
        table.delete().execute()  # pylint: disable=no-value-for-parameter
